import requests
import cloudscraper
import pickle
from pathlib import Path
from .datacamp_utils import Datacamp
from .constants import SESSION_FILE
import os


class Session:
    def __init__(self) -> None:
        self.savefile = Path(SESSION_FILE)
        self.datacamp = Datacamp(self)

    def save(self):
        pickled = pickle.dumps(self)
        self.savefile.write_bytes(pickled)

    @staticmethod
    def load():
        savefile = Path(SESSION_FILE)
        if savefile.exists():
            return pickle.load(savefile.open("rb"))
        return Session()

    def restart(self):
        headers = {
            "Accept-Encoding": "gzip, deflate, sdch",
            "Accept-Language": "en-US,en;q=0.8",
            "Upgrade-Insecure-Requests": "1",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Cache-Control": "max-age=0",
            "Connection": "keep-alive",
            "Referer": "https://learn.datacamp.com/",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:85.0) Gecko/20100101 Firefox/85.0",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate, br",
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "same-site",
            "DNT": "1",
            "Sec-GPC": "1",
            "TE": "Trailers",
        }
        s = requests.Session()
        s.headers = headers
        self.session = cloudscraper.create_scraper(s)
        # remove old session
        try:
            os.remove(self.savefile)
        except:
            pass
        return self

    def get(self, *args, **kwargs):
        return self.session.get(*args, **kwargs)

    def post(self, *args, **kwargs):
        return self.session.post(*args, **kwargs)

    def add_cookie(self, cookie: dict):
        self.session.cookies.set(**cookie)
        return self



from typing import List
from .course import Course


class Track:
    id: int
    title: str
    link: str
    courses: List[Course]

    def __init__(self, id: int, title: str, link: str) -> None:
        self.id = id
        self.title = title
        self.link = link
        self.courses = []



import tempfile

LOGIN_URL = "https://www.datacamp.com/users/sign_in"
LOGIN_DETAILS_URL = "https://www.datacamp.com/api/users/signed_in"

SESSION_FILE = tempfile.gettempdir() + "/.datacamp"

PROFILE_URL = "https://www.datacamp.com/profile/{slug}"
COURSE_DETAILS_API = "https://campus-api.datacamp.com/api/courses/{id}/"
EXERCISE_DETAILS_API = "https://campus-api.datacamp.com/api/exercise/{id}"
VIDEO_DETAILS_API = "https://projector.datacamp.com/api/videos/{hash}"
PROGRESS_API = "https://campus-api.datacamp.com/api/courses/{course_id}/chapters/{chapter_id}/progress"

LANGMAP = {
    "en": "English",
    "zh": "Chinese simplified",
    "fr": "French",
    "de": "German",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "pt": "Portuguese",
    "ru": "Russian",
    "es": "Spanish",
}



# Generated by https://quicktype.io

from typing import List, Any, Optional
from .course import TypeEnum

import tomd


class Data:
    id: int
    type: str
    assignment: Optional[str]
    title: Optional[str]
    sample_code: str
    instructions: Optional[str]
    number: int
    sct: str
    pre_exercise_code: str
    solution: str
    hint: Optional[str]
    attachments: None
    xp: int
    possible_answers: List[Any]
    feedbacks: List[Any]
    question: str
    subexercises: Optional[List["Data"]]
    course_id: Optional[int]
    chapter_id: Optional[int]
    runtime_config: Optional[str]
    language: Optional[str]

    def __init__(
        self,
        id: int,
        type: str,
        assignment: Optional[str] = None,
        title: Optional[str] = None,
        number: int = None,
        hint: Optional[str] = None,
        xp: int = None,
        possible_answers: List[Any] = None,
        feedbacks: List[Any] = None,
        course_id: Optional[int] = None,
        chapter_id: Optional[int] = None,
        runtime_config: Optional[str] = None,
        language: Optional[str] = None,
        subexercises: Optional[List["Data"]] = None,
        instructions: Optional[str] = None,
        attachments: None = None,
        sample_code: str = None,
        pre_exercise_code: str = None,
        solution: str = None,
        sct: str = None,
        question: str = None,
        **kwargs,
    ) -> None:
        self.id = id
        self.type = type
        self.assignment = assignment
        self.title = title
        self.sample_code = sample_code
        self.instructions = instructions
        self.number = number
        self.sct = sct
        self.pre_exercise_code = pre_exercise_code
        self.solution = solution
        self.hint = hint
        self.attachments = attachments
        self.xp = xp
        self.possible_answers = possible_answers
        self.feedbacks = feedbacks
        self.question = question
        self.subexercises = (
            [e.get("id") for e in subexercises] if subexercises else None
        )
        self.course_id = course_id
        self.chapter_id = chapter_id
        self.runtime_config = runtime_config
        self.language = language


class Exercise:
    data: Any
    id: int
    type: str
    version: str
    last_attempt: Optional[str]

    def __init__(
        self,
        data: Data,
        id: int,
        type: str,
        version: str,
        last_attempt: str = None,
        **kwargs,
    ) -> None:
        self.id = id
        self.type = type
        self.version = version
        if not self.is_video:
            self.data = Data(**data)
        else:
            self.data = data
        self.last_attempt = last_attempt

    @property
    def is_video(self):
        return self.type == TypeEnum.VIDEO_EXERCISE.value

    @property
    def is_python(self):
        return self.data.language == "python"

    def __str__(self) -> str:
        html = (
            "<h1> {}</h1>\n<pre><code>Exercise ID {}</code></pre>\n<h2> Assignment </h2>{}\n".format(
                self.data.title, self.id, self.data.assignment
            )
            + self.get_pre_exercise_code()
            + self.get_instructions()
            + self.get_sample_code()
            + self.get_anwsers()
            + self.get_hints()
            + self.get_solution()
        )
        return tomd.convert(html)

    def get_hints(self):
        code = "<h2> Hints </h2> {}<p></p>"
        if self.data.hint:
            return code.format(self.data.hint)
        return ""

    def get_anwsers(self):
        code = "<h2> Answers </h2>{}<p></p>"
        if self.data.possible_answers:
            return code.format(self._get_ordered_list(self.data.possible_answers))
        # return code.format("No answers were found.")
        return ""

    def get_instructions(self):
        code = "<h2> Instructions </h2>{}<p></p>"
        if self.data.instructions:
            return code.format(self.data.instructions)
        return ""

    def _get_ordered_list(self, list):
        return "<ol>{}</ol>".format("\n".join(f"<li>{i}</li>" for i in list))

    def get_solution(self):
        code = "<h2> Solution </h2>{}<p></p>"
        if self.data.feedbacks:
            return code.format(self._get_ordered_list(self.data.feedbacks))
        if self.data.solution:
            return code.format(self._get_code(self.data.solution))
        return code.format("<p>No solution was found.</p>")

    def get_sample_code(self):
        if self.data.sample_code:
            return self._get_code(self.data.sample_code)
        return ""

    def get_pre_exercise_code(self):
        code = "<h2> Pre exercise code </h2> {}<p></p>"
        if self.data.pre_exercise_code:
            return code.format(self._get_code(self.data.pre_exercise_code))
        return ""

    def _get_code(self, code):
        return f"<pre><code>{code}</code></pre>"


from enum import Enum


class Language(str, Enum):
    EN = "en"
    ZH = "zh"
    FR = "fr"
    DE = "de"
    IT = "it"
    JA = "ja"
    KO = "ko"
    PT = "pt"
    RU = "ru"
    ES = "es"
    NONE = "none"


from .session import Session
import copyreg
import ssl
from colorama import init

# use Colorama to make Termcolor work on Windows too
init()

active_session = Session.load()
datacamp = active_session.datacamp
# handle ssl saving issues
copyreg.pickle(ssl.SSLContext, lambda obj: (obj.__class__, (obj.protocol,)))


# Generated by https://quicktype.io

from enum import Enum
from typing import Optional, List, Any


class TypeEnum(Enum):
    FINAL_SLIDE = "FinalSlide"
    FULL_SLIDE = "FullSlide"
    TITLE_SLIDE = "TitleSlide"


class Structure:
    number: int
    type: TypeEnum
    key: str
    script: str
    title: str
    instructor_name: Optional[str]
    instructor_title: Optional[str]
    technology: Optional[str]
    citations: List[Any]
    code_zoom: int
    disable_transition: bool
    hide_slide_in_video: bool
    hide_title: bool
    use_full_width: bool
    part1: Optional[str]

    def __init__(
        self,
        number: int,
        type: TypeEnum,
        key: str,
        script: str,
        title: str,
        instructor_name: Optional[str],
        instructor_title: Optional[str],
        technology: Optional[str],
        citations: List[Any],
        code_zoom: int,
        disable_transition: bool,
        hide_slide_in_video: bool,
        hide_title: bool,
        use_full_width: bool,
        part1: Optional[str],
        **kwargs
    ) -> None:
        self.number = number
        self.type = type
        self.key = key
        self.script = script
        self.title = title
        self.instructor_name = instructor_name
        self.instructor_title = instructor_title
        self.technology = technology
        self.citations = citations
        self.code_zoom = code_zoom
        self.disable_transition = disable_transition
        self.hide_slide_in_video = hide_slide_in_video
        self.hide_title = hide_title
        self.use_full_width = use_full_width
        self.part1 = part1


class SlideDeck:
    key: str
    plain_video_hls_link: str
    plain_video_mp4_link: str
    plain_video_raw_link: None
    structure: List[Structure]
    timings: str
    title: str
    transformations: str

    def __init__(
        self,
        key: str,
        plain_video_hls_link: str,
        plain_video_mp4_link: str,
        plain_video_raw_link: None,
        structure: List[Structure],
        timings: str,
        title: str,
        transformations: str,
    ) -> None:
        self.key = key
        self.plain_video_hls_link = plain_video_hls_link
        self.plain_video_mp4_link = plain_video_mp4_link
        self.plain_video_raw_link = plain_video_raw_link
        self.structure = [Structure(**s) for s in structure]
        self.timings = timings
        self.title = title
        self.transformations = transformations


class Subtitle:
    language: str
    link: str

    def __init__(self, language: str, link: str) -> None:
        self.language = language
        self.link = link


class Video:
    audio_link: str
    key: str
    render_dynamically: int
    script_link: str
    slide_deck: SlideDeck
    slides_link: str
    subtitle_vtt_link: str
    subtitles: List[Subtitle]
    thumbnail_link: None
    transcript_timings: None
    type: str
    video_hls_link: None
    video_mp4_link: str
    video_raw_link: None

    def __init__(
        self,
        audio_link: str,
        key: str,
        render_dynamically: int,
        script_link: str,
        slide_deck: SlideDeck,
        slides_link: str,
        subtitle_vtt_link: str,
        subtitles: List[Subtitle],
        thumbnail_link: None,
        transcript_timings: None,
        type: str,
        video_hls_link: None,
        video_mp4_link: str,
        video_raw_link: None,
    ) -> None:
        self.audio_link = audio_link
        self.key = key
        self.render_dynamically = render_dynamically
        self.script_link = script_link
        self.slide_deck = slide_deck
        self.slides_link = slides_link
        self.subtitle_vtt_link = subtitle_vtt_link
        self.subtitles = [Subtitle(**s) for s in subtitles]
        self.thumbnail_link = thumbnail_link
        self.transcript_timings = transcript_timings
        self.type = type
        self.video_hls_link = video_hls_link
        self.video_mp4_link = video_mp4_link
        self.video_raw_link = video_raw_link



from pathlib import Path
from typing import List, Optional
import typer

from .templates.lang import Language

from .helper import Logger
from . import datacamp, active_session
import os


__version__ = "2.1.0"


def version_callback(value: bool):
    if value:
        typer.echo(f"Datacamp Downloader CLI Version: {__version__}")
        raise typer.Exit()


def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version.",
    ),
):
    pass


app = typer.Typer(callback=main)


@app.command()
def login(
    username: str = typer.Option(..., "-u", "--username", prompt=True),
    password: str = typer.Option(..., "-p", "--password", prompt=True, hide_input=True),
):
    """Log in to Datacamp using your username and password."""
    datacamp.login(username, password)


@app.command()
def set_token(token: str = typer.Argument(...)):
    """Log in to Datacamp using your token."""
    datacamp.set_token(token)


@app.command()
def tracks(
    refresh: Optional[bool] = typer.Option(
        False, "--refresh", "-r", is_flag=True, help="Refresh completed tracks."
    )
):
    """List your completed tracks."""
    datacamp.list_completed_tracks(refresh)


@app.command()
def courses(
    refresh: Optional[bool] = typer.Option(
        False, "--refresh", "-r", is_flag=True, help="Refresh completed courses."
    )
):
    """List your completed courses."""
    datacamp.list_completed_courses(refresh)


@app.command()
def download(
    ids: List[str] = typer.Argument(
        ...,
        help="IDs for courses/tracks to download or `all` to download all your completed courses or `all-t` to download all your completed tracks.",
    ),
    path: Path = typer.Option(
        Path(os.getcwd() + "/Datacamp"),
        "--path",
        "-p",
        help="Path to the download directory.",
        dir_okay=True,
        file_okay=False,
    ),
    slides: Optional[bool] = typer.Option(
        True,
        "--slides/--no-slides",
        help="Download slides.",
    ),
    datasets: Optional[bool] = typer.Option(
        True,
        "--datasets/--no-datasets",
        help="Download datasets.",
    ),
    videos: Optional[bool] = typer.Option(
        True,
        "--videos/--no-videos",
        help="Download videos.",
    ),
    exercises: Optional[bool] = typer.Option(
        True,
        "--exercises/--no-exercises",
        help="Download exercises.",
    ),
    subtitles: Optional[List[Language]] = typer.Option(
        [Language.EN.value],
        "--subtitles",
        "-st",
        help="Choose subtitles to download.",
        case_sensitive=False,
    ),
    audios: Optional[bool] = typer.Option(
        False,
        "--audios/--no-audios",
        help="Download audio files.",
    ),
    scripts: Optional[bool] = typer.Option(
        True,
        "--scripts/--no-scripts",
        "--transcript/--no-transcript",
        show_default=True,
        help="Download scripts or transcripts.",
    ),
    python_file: Optional[bool] = typer.Option(
        True,
        "--python-file/--no-python-file",
        show_default=True,
        help="Download your own solution as a python file if available.",
    ),
    warnings: Optional[bool] = typer.Option(
        True,
        "--no-warnings",
        flag_value=False,
        is_flag=True,
        help="Disable warnings.",
    ),
    overwrite: Optional[bool] = typer.Option(
        False,
        "--overwrite",
        "-w",
        flag_value=True,
        is_flag=True,
        help="Overwrite files if exist.",
    ),
):
    """Download courses/tracks given their ids.

    Example: `datacamp download id1 id2 id3`\n
    To download all your completed courses run:
    \t`datacamp download all`\n
    To download all your completed tracks run:
    \t`datacamp download all-t`
    """
    Logger.show_warnings = warnings
    datacamp.download(
        ids,
        path,
        slides=slides,
        datasets=datasets,
        videos=videos,
        exercises=exercises,
        subtitles=subtitles,
        audios=audios,
        scripts=scripts,
        overwrite=overwrite,
        last_attempt=python_file,
    )


@app.command()
def reset():
    """Restart the session."""
    active_session.restart()


from pathlib import Path
import sys
import time
import threading
import itertools
import re

from texttable import Texttable
from termcolor import colored


class Logger:
    show_warnings = True

    @classmethod
    def error(cls, text):
        Logger.print(text, "ERROR:", "red")

    @classmethod
    def clear(cls):
        sys.stdout.write("\r" + " " * 100 + "\r")

    @classmethod
    def warning(cls, text):
        if cls.show_warnings:
            Logger.print(text, "WARNING:", "yellow")

    @classmethod
    def info(cls, text):
        Logger.print(text, "INFO:", "green")

    @classmethod
    def print(cls, text, head, color=None, background=None, end="\n"):
        Logger.clear()
        print(colored(f"{head}", color, background), text, end=end)

    @classmethod
    def print_table(cls, rows):
        Logger.clear()
        table = Texttable()
        table.set_max_width(100)
        table.add_rows(rows)
        print(table.draw())


def animate_wait(f):
    done = False

    def animate():
        for c in itertools.cycle(list("/—\|")):
            if done:
                sys.stdout.write("\r")
                break
            sys.stdout.write("\rPlease wait " + c)
            time.sleep(0.1)
            sys.stdout.flush()

    def wrapper(*args):
        nonlocal done
        done = False
        t = threading.Thread(target=animate)
        t.start()
        output = f(*args)
        done = True
        return output

    return wrapper


def correct_path(path: str):
    return re.sub("[^-a-zA-Z0-9_.() /]+", "", path)


def download_file(
    session, link: str, path: Path, progress=True, max_retry=10, overwrite=False
):
    # start = time.clock()
    if not overwrite and path.exists():
        Logger.warning(f"{path.absolute()} is already downloaded")
        return

    for i in range(max_retry):
        try:
            response = session.get(link, stream=True)
            i = -1
            break
        except Exception:
            Logger.print(f"", f"Retry [{i+1}/{max_retry}]", "magenta", end="")

    if i != -1:
        Logger.error(f"Failed to download {link}")
        return

    path.parent.mkdir(exist_ok=True, parents=True)
    total_length = response.headers.get("content-length")

    with path.open("wb") as f:
        if total_length is None:  # no content length header
            f.write(response.content)
        else:
            dl = 0
            total_length = int(total_length)
            for data in response.iter_content(chunk_size=1024 * 1024):  # 1MB
                dl += len(data)
                f.write(data)
                if progress:
                    print_progress(dl, total_length, path.name)
    if progress:
        sys.stdout.write("\n")


def print_progress(progress, total, name, max=50):
    done = int(max * progress / total)
    Logger.print(
        "[%s%s] %d%%" % ("=" * done, " " * (max - done), done * 2),
        f"Downloading [{name}]",
        "blue",
        end="\r",
    )
    sys.stdout.flush()


def save_text(path: Path, content: str, overwrite=False):
    if not path.is_file:
        Logger.error(f"{path.absolute()} isn't a file")
        return
    if not overwrite and path.exists():
        Logger.warning(f"{path.absolute()} is already downloaded")
        return
    path.parent.mkdir(exist_ok=True, parents=True)
    path.write_text(content, encoding="utf8")
    # Logger.info(f"{path.name} has been saved.")


def fix_track_link(link):
    if "?" in link:
        link += "&embedded=true"
    else:
        link += "?embedded=true"
    return link


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    required = fh.read().splitlines()

setup(
    name="datacamp-downloader",
    version="2.1",
    author="Mohammad Al-Fetyani",
    author_email="m4bh@hotmail.com",
    description="Download your completed courses on Datacamp easily!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TRoboto/datacamp-downloader",
    project_urls={
        "Bug Tracker": "https://github.com/TRoboto/datacamp-downloader/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    install_requires=required,
    setup_requires=["setuptools-git"],
    packages=find_packages(where="src"),
    include_package_data=True,
    # package_data={"datacamp_downloader": ["*", "*/*", "*/*/*"]},
    python_requires=">=3.6",
    entry_points={"console_scripts": ["datacamp=datacamp_downloader.downloader:app"]},
)


# Generated by https://quicktype.io

from enum import Enum
from typing import Optional, List, Any


class TypeEnum(Enum):
    MULTIPLE_CHOICE_EXERCISE = "MultipleChoiceExercise"
    NORMAL_EXERCISE = "NormalExercise"
    VIDEO_EXERCISE = "VideoExercise"


class Exercise:
    type: TypeEnum
    title: str
    aggregate_xp: int
    number: int
    url: str

    def __init__(
        self,
        type: TypeEnum,
        title: str,
        aggregate_xp: int,
        number: int,
        url: str,
        **kwargs
    ) -> None:
        self.type = type
        self.title = title
        self.aggregate_xp = aggregate_xp
        self.number = number
        self.url = url


class Chapter:
    id: int
    title_meta: str
    title: str
    description: str
    number: int
    slug: str
    nb_exercises: int
    badge_completed_url: str
    badge_uncompleted_url: str
    last_updated_on: str
    slides_link: str
    free_preview: Optional[bool]
    xp: int
    number_of_videos: int
    exercises: List[Exercise]

    def __init__(
        self,
        id: int,
        title_meta: str,
        title: str,
        description: str,
        number: int,
        slug: str,
        nb_exercises: int,
        badge_completed_url: str,
        badge_uncompleted_url: str,
        last_updated_on: str,
        slides_link: str,
        free_preview: Optional[bool],
        xp: int,
        number_of_videos: int,
        exercises: List[Exercise],
        **kwargs
    ) -> None:
        self.id = id
        self.title_meta = title_meta
        self.title = title
        self.description = description
        self.number = number
        self.slug = slug
        self.nb_exercises = nb_exercises
        self.badge_completed_url = badge_completed_url
        self.badge_uncompleted_url = badge_uncompleted_url
        self.last_updated_on = last_updated_on
        self.slides_link = slides_link
        self.free_preview = free_preview
        self.xp = xp
        self.number_of_videos = number_of_videos
        self.exercises = [Exercise(**c) for c in exercises]


class Collaborator:
    avatar_url: str
    full_name: str

    def __init__(self, avatar_url: str, full_name: str) -> None:
        self.avatar_url = avatar_url
        self.full_name = full_name


class Dataset:
    asset_url: str
    name: str

    def __init__(self, asset_url: str, name: str) -> None:
        self.asset_url = asset_url
        self.name = name


class Instructor:
    id: int
    marketing_biography: str
    biography: str
    avatar_url: str
    full_name: str
    instructor_path: str

    def __init__(
        self,
        id: int,
        marketing_biography: str,
        biography: str,
        avatar_url: str,
        full_name: str,
        instructor_path: str,
        **kwargs
    ) -> None:
        self.id = id
        self.marketing_biography = marketing_biography
        self.biography = biography
        self.avatar_url = avatar_url
        self.full_name = full_name
        self.instructor_path = instructor_path


class SharingLinks:
    twitter: str
    facebook: str

    def __init__(self, twitter: str, facebook: str) -> None:
        self.twitter = twitter
        self.facebook = facebook


class Track:
    path: str
    title_with_subtitle: str

    def __init__(self, path: str, title_with_subtitle: str) -> None:
        self.path = path
        self.title_with_subtitle = title_with_subtitle


class Course:
    id: int
    title: str
    description: str
    short_description: str
    author_field: None
    author_bio: None
    author_image: str
    nb_of_subscriptions: int
    slug: str
    image_url: str
    image_thumbnail_url: str
    last_updated_on: str
    link: str
    should_cache: bool
    type: str
    difficulty_level: int
    state: str
    university: None
    sharing_links: SharingLinks
    marketing_video: str
    programming_language: str
    paid: bool
    time_needed: None
    xp: int
    topic_id: int
    technology_id: int
    reduced_outline: None
    runtime_config: None
    lti_only: bool
    instructors: List[Instructor]
    collaborators: List[Collaborator]
    datasets: List[Dataset]
    tracks: List[Track]
    prerequisites: List[Any]
    time_needed_in_hours: int
    seo_title: str
    seo_description: str
    archived_at: None
    original_image_url: str
    external_slug: str
    chapters: List[Chapter]

    def __init__(
        self,
        id: int,
        title: str,
        description: str,
        short_description: str,
        author_field: None,
        author_bio: None,
        author_image: str,
        nb_of_subscriptions: int,
        slug: str,
        image_url: str,
        image_thumbnail_url: str,
        last_updated_on: str,
        link: str,
        should_cache: bool,
        type: str,
        difficulty_level: int,
        state: str,
        university: None,
        sharing_links: SharingLinks,
        marketing_video: str,
        programming_language: str,
        paid: bool,
        time_needed: None,
        xp: int,
        topic_id: int,
        technology_id: int,
        reduced_outline: None,
        runtime_config: None,
        lti_only: bool,
        instructors: List[Instructor],
        collaborators: List[Collaborator],
        datasets: List[Dataset],
        tracks: List[Track],
        prerequisites: List[Any],
        time_needed_in_hours: int,
        seo_title: str,
        seo_description: str,
        archived_at: None,
        original_image_url: str,
        external_slug: str,
        chapters: List[Chapter],
        **kwargs
    ) -> None:
        self.id = id
        self.title = title
        self.description = description
        self.short_description = short_description
        self.author_field = author_field
        self.author_bio = author_bio
        self.author_image = author_image
        self.nb_of_subscriptions = nb_of_subscriptions
        self.slug = slug
        self.image_url = image_url
        self.image_thumbnail_url = image_thumbnail_url
        self.last_updated_on = last_updated_on
        self.link = link
        self.should_cache = should_cache
        self.type = type
        self.difficulty_level = difficulty_level
        self.state = state
        self.university = university
        self.sharing_links = sharing_links
        self.marketing_video = marketing_video
        self.programming_language = programming_language
        self.paid = paid
        self.time_needed = time_needed
        self.xp = xp
        self.topic_id = topic_id
        self.technology_id = technology_id
        self.reduced_outline = reduced_outline
        self.runtime_config = runtime_config
        self.lti_only = lti_only
        self.instructors = [Instructor(**c) for c in instructors]
        self.collaborators = [Collaborator(**c) for c in collaborators]
        self.datasets = [Dataset(**c) for c in datasets]
        self.tracks = [Track(**c) for c in tracks]
        self.prerequisites = prerequisites
        self.time_needed_in_hours = time_needed_in_hours
        self.seo_title = seo_title
        self.seo_description = seo_description
        self.archived_at = archived_at
        self.original_image_url = original_image_url
        self.external_slug = external_slug
        self.chapters = [Chapter(**c) for c in chapters]



from pathlib import Path
import sys
from bs4 import BeautifulSoup
import re

from .constants import (
    COURSE_DETAILS_API,
    EXERCISE_DETAILS_API,
    LANGMAP,
    LOGIN_DETAILS_URL,
    LOGIN_URL,
    PROFILE_URL,
    PROGRESS_API,
    VIDEO_DETAILS_API,
)
from .helper import (
    Logger,
    animate_wait,
    correct_path,
    download_file,
    fix_track_link,
    print_progress,
    save_text,
)
from .templates.track import Track
from .templates.course import Chapter, Course
from .templates.exercise import Exercise
from .templates.video import Video


def login_required(f):
    def wrapper(*args, **kwargs):
        self = args[0]
        if not isinstance(self, Datacamp):
            Logger.error(f"{login_required.__name__} can only decorate Datacamp class.")
            return
        if not self.loggedin:
            Logger.error("Login first!")
            return
        if not self.has_active_subscription:
            Logger.error("No active subscription found.")
            return
        return f(*args, **kwargs)

    return wrapper


def try_except_request(f):
    def wrapper(*args, **kwargs):
        self = args[0]
        if not isinstance(self, Datacamp):
            Logger.error(
                f"{try_except_request.__name__} can only decorate Datacamp class."
            )
            return

        try:
            return f(*args, **kwargs)
        except Exception:
            Logger.warning(f"Couldn't run {f.__name__} with inputs {args[1:]}")
        return

    return wrapper


class Datacamp:
    def __init__(self, session) -> None:

        self.session = session
        self.username = None
        self.password = None
        self.token = None
        self.has_active_subscription = False
        self.loggedin = False
        self.login_data = None

        self.courses = []
        self.tracks = []

        self.not_found_courses = set()

    @animate_wait
    @try_except_request
    def login(self, username, password):
        if username == self.username and self.password == password and self.loggedin:
            Logger.info("Already logged in!")
            return

        self.session.restart()
        self.username = username
        self.password = password

        req = self.session.get(LOGIN_URL)
        if not req or req.status_code != 200 or not req.text:
            Logger.error("Cannot access datacamp website!")
            return
        soup = BeautifulSoup(req.text, "html.parser")
        authenticity_token = soup.find("input", {"name": "authenticity_token"}).get(
            "value"
        )
        if not authenticity_token:
            Logger.error("Cannot find authenticity_token attribute!")
            return

        login_req = self.session.post(
            LOGIN_URL,
            params=[
                ("authenticity_token", authenticity_token),
                ("user[email]", username),
                ("user[password]", password),
            ],
        )
        if (
            not login_req
            or login_req.status_code != 200
            or "/users/sign_up" in login_req.text
        ):
            Logger.error("Incorrent username/password")
            return

        self._set_profile()

    @animate_wait
    def set_token(self, token):
        if self.token == token and self.loggedin:
            Logger.info("Already logged in!")
            return

        self.session.restart()
        self.token = token
        cookie = {
            "name": "_dct",
            "value": token,
            "domain": ".datacamp.com",
            "secure": True,
        }
        self.session.add_cookie(cookie)
        self._set_profile()

    @login_required
    @animate_wait
    def list_completed_tracks(self, refresh):
        if refresh or not self.tracks:
            self.get_completed_tracks()
        rows = [["#", "ID", "Title", "Courses"]]
        for i, track in enumerate(self.tracks, 1):
            rows.append([i, track.id, track.title, len(track.courses)])
        Logger.print_table(rows)

    @login_required
    @animate_wait
    def list_completed_courses(self, refresh):
        if refresh or not self.courses:
            self.get_completed_courses()
        rows = [["#", "ID", "Title", "Datasets", "Exercises", "Videos"]]
        for i, course in enumerate(self.courses, 1):
            all_exercises_count = sum([c.nb_exercises for c in course.chapters])
            videos_count = sum([c.number_of_videos for c in course.chapters])
            rows.append(
                [
                    i,
                    course.id,
                    course.title,
                    len(course.datasets),
                    all_exercises_count - videos_count,
                    videos_count,
                ]
            )
        Logger.print_table(rows)

    @login_required
    def download(self, ids, directory, **kwargs):
        self.overwrite = kwargs.get("overwrite")
        if "all-t" in ids:
            if not self.tracks:
                animate_wait(self.get_completed_tracks)()
            to_download = self.tracks
        elif "all" in ids:
            if not self.courses:
                animate_wait(self.get_completed_courses)()
            to_download = self.courses
        else:
            to_download = []
            for id in ids:
                if "t" in id:
                    to_download.append(animate_wait(self.get_track)(id))
                elif id.isnumeric():
                    to_download.append(animate_wait(self.get_course)(id))
        # filter none
        to_download = [i for i in to_download if i]

        path = Path(directory) if not isinstance(directory, Path) else directory

        for i, material in enumerate(to_download, 1):
            if not material:
                continue
            Logger.info(
                f"[{i}/{len(to_download)}] Start to download ({material.id}) {material.title}"
            )
            if isinstance(material, Course):
                self.download_course(material, path, **kwargs)
            else:
                self.download_track(material, path, **kwargs)

    def download_normal_exercise(
        self, exercise: Exercise, path: Path, include_last_attempt: bool = False
    ):
        save_text(path, str(exercise), self.overwrite)
        if include_last_attempt and exercise.is_python and exercise.last_attempt:
            save_text(
                path.parent / (path.name[:-3] + f".py"),
                exercise.last_attempt,
                self.overwrite,
            )
        subexs = exercise.data.subexercises
        if subexs:
            for i, subexercise in enumerate(subexs, 1):
                exercise = self._get_exercise(subexercise)
                self.download_normal_exercise(
                    exercise, path.parent / (path.name[:-3] + f"_sub{i}.md")
                )

    def download_track(self, track: Track, path: Path, **kwargs):
        path = path / correct_path(track.title)
        for i, course in enumerate(track.courses, 1):
            Logger.info(
                f"[{i}/{len(track.courses)}] Download ({course.id}) {course.title} from ({track.title} Track)"
            )
            self.download_course(course, path, f"{i}-", **kwargs)

    def download_course(self, course: Course, path: Path, index="", **kwargs):
        download_path = path / (
            index + correct_path(course.slug or course.title.lower().replace(" ", "-"))
        )
        if kwargs.get("datasets") and course.datasets:
            for i, dataset in enumerate(course.datasets, 1):
                print_progress(i, len(course.datasets), f"datasets")
                if dataset.asset_url:
                    download_file(
                        self.session,
                        dataset.asset_url,
                        download_path
                        / "datasets"
                        / correct_path(dataset.asset_url.split("/")[-1]),
                        False,
                        overwrite=self.overwrite,
                    )
            sys.stdout.write("\n")
        for chapter in course.chapters:
            cpath = download_path / self._get_chapter_name(chapter)
            if kwargs.get("slides") and chapter.slides_link:
                download_file(
                    self.session,
                    chapter.slides_link,
                    cpath / correct_path(chapter.slides_link.split("/")[-1]),
                    overwrite=self.overwrite,
                )
            if (
                kwargs.get("exercises")
                or kwargs.get("videos")
                or kwargs.get("audios")
                or kwargs.get("scripts")
            ):
                self.download_others(course.id, chapter, cpath, **kwargs)

    def download_others(self, course_id, chapter: Chapter, path: Path, **kwargs):
        exercises = kwargs.get("exercises")
        videos = kwargs.get("videos")
        audios = kwargs.get("audios")
        scripts = kwargs.get("scripts")
        subtitles = kwargs.get("subtitles")
        last_attempt = kwargs.get("last_attempt")
        ids = self._get_exercises_ids(course_id, chapter.id)
        last_attempts = self.get_exercises_last_attempt(course_id, chapter.id)
        exercise_counter = 1
        video_counter = 1
        for i, id in enumerate(ids, 1):
            print_progress(i, len(ids), f"chapter {chapter.number}")
            exercise = self._get_exercise(id)
            exercise.last_attempt = last_attempts[id]
            if not exercise:
                continue
            if exercises and not exercise.is_video:
                self.download_normal_exercise(
                    exercise,
                    path / "exercises" / f"ex{exercise_counter}.md",
                    last_attempt,
                )
                exercise_counter += 1
            if exercise.is_video:
                video = self._get_video(exercise.data.get("projector_key"))
                video_path = path / "videos" / f"ch{chapter.number}_{video_counter}"
                if videos and video.video_mp4_link:
                    download_file(
                        self.session,
                        video.video_mp4_link,
                        video_path.with_suffix(".mp4"),
                        overwrite=self.overwrite,
                    )
                if audios and video.audio_link:
                    download_file(
                        self.session,
                        video.audio_link,
                        path / "audios" / f"ch{chapter.number}_{video_counter}.mp3",
                        False,
                        overwrite=self.overwrite,
                    )
                if scripts and video.script_link:
                    download_file(
                        self.session,
                        video.script_link,
                        path / "scripts" / (video_path.name + "_script.md"),
                        False,
                        overwrite=self.overwrite,
                    )
                if subtitles and video.subtitles:
                    for sub in subtitles:
                        subtitle = self._get_subtitle(sub, video)
                        if not subtitle:
                            continue
                        download_file(
                            self.session,
                            subtitle.link,
                            video_path.parent / (video_path.name + f"_{sub}.vtt"),
                            False,
                            overwrite=self.overwrite,
                        )
                video_counter += 1
            print_progress(i, len(ids), f"chapter {chapter.number}")
        sys.stdout.write("\n")

    def get_completed_tracks(self):
        self.tracks = []
        profile = self.session.get(PROFILE_URL.format(slug=self.login_data["slug"]))
        soup = BeautifulSoup(profile.text, "html.parser")
        tracks_title = soup.findAll("div", {"class": "track-block__main"})
        tracks_link = soup.findAll(
            "a", {"href": re.compile("^/tracks"), "class": "shim"}
        )
        for i in range(len(tracks_link)):
            link = "https://www.datacamp.com" + tracks_link[i]["href"]
            self.tracks.append(
                Track(
                    f"t{i + 1}",
                    tracks_title[i].getText().replace("\n", " ").strip(),
                    link,
                )
            )
        courses = set()
        # add courses
        for track in self.tracks:
            track.courses = self._get_courses_from_link(fix_track_link(track.link))
            courses.update(track.courses)
        # add to courses
        current_ids = [c.id for c in self.courses]
        for course in courses:
            if course.id not in current_ids:
                self.courses.append(course)

        self.session.save()
        return self.tracks

    def get_completed_courses(self):
        self.courses = self._get_courses_from_link(
            PROFILE_URL.format(slug=self.login_data["slug"])
        )

        self.session.save()
        return self.courses

    def get_course(self, id):
        if id in self.not_found_courses:
            return
        for course in self.courses:
            if course.id == id:
                return course
        return self._get_course(id)

    @try_except_request
    def get_exercises_last_attempt(self, course_id, chapter_id):
        res = self.session.get(
            PROGRESS_API.format(course_id=course_id, chapter_id=chapter_id)
        )
        data = res.json()
        if "error" in data:
            raise ValueError("Cannot get info.")
        last_attempt = {e["exercise_id"]: e["last_attempt"] for e in data}
        return last_attempt

    def get_track(self, id):
        if not self.tracks:
            self.get_completed_tracks()
        for track in self.tracks:
            if track.id == id:
                return track

    @try_except_request
    def _get_courses_from_link(self, link: str):
        html = self.session.get(link)
        soup = BeautifulSoup(html.text, "html.parser")
        courses_ids = soup.findAll("article", {"class": re.compile("^js-async")})
        courses = []
        for id_tag in courses_ids:
            id = id_tag.get("data-id")
            if not id:
                continue
            course = self.get_course(int(id))
            if course:
                courses.append(course)
        return courses

    def _get_chapter_name(self, chapter: Chapter):
        if chapter.title and chapter.title_meta:
            return correct_path(chapter.slug)
        if chapter.title:
            return correct_path(
                f"chapter-{chapter.number}-{chapter.title.replace(' ', '-').lower()}"
            )
        return f"chapter-{chapter.number}"

    def _set_profile(self):
        page = self.session.get(LOGIN_DETAILS_URL)
        try:
            data = page.json()
        except:
            Logger.error("Incorrect input token!")
            return
        Logger.info("Hi, " + (data["first_name"] or data["last_name"] or data["email"]))

        if data["has_active_subscription"]:
            Logger.info("Active subscription found")
        else:
            Logger.warning("No active subscription found")

        self.loggedin = True
        self.login_data = data
        self.has_active_subscription = data["has_active_subscription"]

        self.session.save()

    def _get_subtitle(self, sub, video: Video):
        if not LANGMAP.get(sub):
            return
        for subtitle in video.subtitles:
            if subtitle.language == LANGMAP[sub]:
                return subtitle

    @try_except_request
    def _get_video(self, id):
        if not id:
            raise ValueError("ID tag not found.")
        res = self.session.get(VIDEO_DETAILS_API.format(hash=id))
        if "error" in res.json():
            raise ValueError("Cannot get info.")
        return Video(**res.json())

    @try_except_request
    def _get_exercises_ids(self, course_id, chapter_id):
        if not course_id or not chapter_id:
            raise ValueError("ID tags not found.")
        res = self.session.get(
            PROGRESS_API.format(course_id=course_id, chapter_id=chapter_id)
        )
        data = res.json()
        if "error" in data:
            raise ValueError("Cannot get info.")
        ids = [e["exercise_id"] for e in data]
        return ids

    @try_except_request
    def _get_exercise(self, id):
        if not id:
            raise ValueError("ID tag not found.")
        res = self.session.get(EXERCISE_DETAILS_API.format(id=id))
        if "error" in res.json():
            raise ValueError("Cannot get info.")
        return Exercise(**res.json())

    @try_except_request
    def _get_course(self, id):
        if not id:
            self.not_found_courses.add(id)
            raise ValueError("ID tag not found.")
        res = self.session.get(COURSE_DETAILS_API.format(id=id))
        if "error" in res.json():
            self.not_found_courses.add(id)
            raise ValueError("Cannot get info.")
        return Course(**res.json())
